
#include "stm32l1xx.h"
#include "hw_config.h"
#include "usb_lib.h"
#include "usb_pwr.h"
#include "dht.h"
#include "command.h"
#include "ti8077_lcd.h"
#include "display.h"
#include "main.h"

#ifdef DEBUG
//#define DEBUG_SWD_PIN  1  /* needs to be set to 1 to enable SWD debug pins, set to 0 for power consumption measurement*/
#endif

RCC_ClocksTypeDef RCC_Clocks;
bool gErrorFlash = false;
int16_t gLastSampleIndex = -1;

/* Private function prototypes -----------------------------------------------*/
void SetSysClock32mhz(void);
void SetSysClock(void);//Low speed clock (4 mhz)

void  RCC_Configuration(void);
void  RTC_Configuration(void);
void  InitUsbPin (void);
void  DisableUsbPin (void);
uint8_t IsUsbPower(void);

void configureWakeup (uint32_t seconds);
void insertionSort(uint16_t *numbers, uint32_t array_size);
uint32_t interquartileMean(uint16_t *array, uint32_t numOfSamples);
void RtcInit(void);
void conf_analog_all_GPIOS(void);
void UsbMain(void);
void DhtMeasureAndDisplay(void);

void ClockQuant(void);

/*******************************************************************************/

/**
  * @brief main entry point.
  * @par Parameters None
  * @retval void None
  * @par Required preconditions: None
  */
                                  
int main(void)
{ 

 /*!< At this stage the microcontroller clock setting is already configured, 
       this is done through SystemInit() function which is called from startup
       file (startup_stm32l1xx_md.s) before to branch to application main.
       To reconfigure the default setting of SystemInit() function, refer to
       system_stm32l1xx.c file
     */ 

  /* Configure Clocks for Application need */
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_LCD | RCC_APB1Periph_PWR, ENABLE);
  //RCC_Configuration();
  //SystemInit();
  
  /* Configure RTC Clocks */
  RTC_Configuration();

  /* Set internal voltage regulator to 1.8V */
  PWR_VoltageScalingConfig(PWR_VoltageScaling_Range1);

  /* Wait Until the Voltage Regulator is ready */
  while (PWR_GetFlagStatus(PWR_FLAG_VOS) != RESET) ;

  /* Enable debug features in low power modes (Sleep, STOP and STANDBY) */
#ifdef  DEBUG_SWD_PIN
  DBGMCU_Config(DBGMCU_SLEEP | DBGMCU_STOP | DBGMCU_STANDBY, ENABLE);
#endif
  
  /* Configure SysTick IRQ and SysTick Timer to generate interrupts */
  RCC_GetClocksFreq(&RCC_Clocks);
  SysTick_Config(RCC_Clocks.HCLK_Frequency / 500);

  DHTInit();
  DHTDelay1ms(2000);//Wait connect programmator
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOB, DISABLE);
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM6, DISABLE);


  conf_analog_all_GPIOS();

  RtcInit();

  LCD_GLASS_Configure_GPIO();
  LCD_GLASS_Init();
  DHTInit();
  //RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOB, DISABLE);
  //RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM6, DISABLE);
  //InitUsbPowerPin();

  // Disable SysTick IRQ and SysTick Timer
  SysTick->CTRL  &= ~ ( SysTick_CTRL_TICKINT_Msk | SysTick_CTRL_ENABLE_Msk );

  DhtMeasureAndDisplay();

  while(1)
  {
    InitUsbPin();
    if(IsUsbPower())
    {
      UsbMain();      
    }

    ClockQuant();

    {
      //Wait XX second
      configureWakeup(10);
      // Enable RTC Wakeup
      RTC_WakeUpCmd(ENABLE);

      // Clear WakeUp flag
      PWR_ClearFlag(PWR_FLAG_WU);
    
      // Enter in wait for interrupt stop mode
      PWR_EnterSTOPMode(PWR_Regulator_LowPower, PWR_STOPEntry_WFI);

      RCC_Configuration();  // reinitialize clock
      
      // After Wake up : Disable Wake up from RTC
      RTC_WakeUpCmd(DISABLE);
    }
  }

}

void DhtMeasureAndDisplay(void)
{
    DHTPowerOn(true);
    DHTDelay1ms(2000);
    DHTReadData();
    DHTPowerOn(false);
    DisableUsbPin();

    Display();
}

void UsbMain(void)
{
  int8_t usb_power_counter = 5;
  DHTDelay1ms(1);
  if(!IsUsbPower())
    return;
  DisplayUsb();
  RCC_DeInit();
  SetSysClock32mhz(); //Hi speed

  SystemCoreClockUpdate();

  UsbInitClockAndSystem();
  while(usb_power_counter>=0)
  {
    if(!IsUsbPower())
    {
      DHTDelay1ms(1);
      usb_power_counter--;
    }

    DHTDelay1ms(1000);
  }

  PowerOff();
  RCC_DeInit();
  SetSysClock();
  Display();
}

/*---------------------------------------------------------------------------*/

void configureWakeup (uint32_t seconds)
{
  /* Declare initialisation structures for (NVIC) and external interupt (EXTI) */
  NVIC_InitTypeDef NVIC_InitStructure;
  EXTI_InitTypeDef EXTI_InitStructure;

  /* Clear IT pending bit from external interrupt Line 20 */
  EXTI_ClearITPendingBit(EXTI_Line20);
  
  /* Initialise EXTI using its init structure */
  EXTI_InitStructure.EXTI_Line = EXTI_Line20;			 // interrupt generated on RTC Wakeup event (Line 20)
  EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;    // Use EXTI line as interrupt
  EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising; // Trigg interrupt on rising edge detection
  EXTI_InitStructure.EXTI_LineCmd = ENABLE;				 // Enable EXTI line
  EXTI_Init(&EXTI_InitStructure);						 
  
  /* Initialise the NVIC interrupts (IRQ) using its init structure */
  NVIC_InitStructure.NVIC_IRQChannel = RTC_WKUP_IRQn;        // set IRQ channel to RTC Wakeup Interrupt  
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;	 // set channel Preemption priority to 0
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;         // set channel sub priority to 0
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;	         // Enable channel
  NVIC_Init(&NVIC_InitStructure);
  
  /* Clear Wake-up flag */  
  PWR->CR |= PWR_CR_CWUF;

  /* Enable PWR clock */
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR, ENABLE); 

  /* Allow access to RTC */
  PWR_RTCAccessCmd(ENABLE);

  /* Enable Low Speed External clock */
  RCC_LSEConfig(RCC_LSE_ON); 

  /* Wait till LSE is ready */
  while (RCC_GetFlagStatus(RCC_FLAG_LSERDY) == RESET);

  /* Select LSE clock as RCC Clock source */
  RCC_RTCCLKConfig(RCC_RTCCLKSource_LSE);

  /* Enable the RTC Clock */
  RCC_RTCCLKCmd(ENABLE);

  /* Wait for RTC APB registers synchronisation */
  RTC_WaitForSynchro();

  /* Select 1Hz clock for RTC wake up*/
  RTC_WakeUpClockConfig(RTC_WakeUpClock_CK_SPRE_16bits);
  
  /* Set Wakeup auto-reload value to 2 sec */
  RTC_SetWakeUpCounter(seconds-1); 

  /* Clear RTC Interrupt pending bit */
  RTC_ClearITPendingBit(RTC_IT_WUT);
  
  /* Clear EXTI line20 Interrupt pending bit */
  EXTI_ClearITPendingBit(EXTI_Line20);

  /* Enable the Wakeup Interrupt */
  RTC_ITConfig(RTC_IT_WUT, ENABLE);
}

/**
  * @brief  Configures the different system clocks.
  * @param  None
  * @retval None
  */
void RCC_Configuration(void)
{  
  
  /* Enable HSI Clock */
  RCC_HSICmd(ENABLE);
  
  /*!< Wait till HSI is ready */
  while (RCC_GetFlagStatus(RCC_FLAG_HSIRDY) == RESET)
  {}

  RCC_SYSCLKConfig(RCC_SYSCLKSource_HSI);
  
  RCC_MSIRangeConfig(RCC_MSIRange_6);

  RCC_HSEConfig(RCC_HSE_OFF);  
  if(RCC_GetFlagStatus(RCC_FLAG_HSERDY) != RESET )
  {
    while(1);
  }
 
  /* Enable  comparator clock LCD and PWR mngt */
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_LCD | RCC_APB1Periph_PWR, ENABLE);
    
  /* Enable  SYSCFG */
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);

  //For USB power pin
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
}


void RTC_Configuration(void)
{
  
/* Allow access to the RTC */
  PWR_RTCAccessCmd(ENABLE);

  /* Reset Backup Domain */
  RCC_RTCResetCmd(ENABLE);
  RCC_RTCResetCmd(DISABLE);

  /* LSE Enable */
  RCC_LSEConfig(RCC_LSE_ON);

  /* Wait till LSE is ready */
  while (RCC_GetFlagStatus(RCC_FLAG_LSERDY) == RESET)
  {}
  
  RCC_RTCCLKCmd(ENABLE);
   
  /* LCD Clock Source Selection */
  RCC_RTCCLKConfig(RCC_RTCCLKSource_LSE);

}

/**
  * @brief  To initialize the I/O ports
  * @caller main
  * @param None
  * @retval None
  */


void conf_analog_all_GPIOS(void)
{
  GPIO_InitTypeDef GPIO_InitStructure;

  /* Enable GPIOs clock */ 	
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA | RCC_AHBPeriph_GPIOB | 
                        RCC_AHBPeriph_GPIOC | RCC_AHBPeriph_GPIOD | 
                        RCC_AHBPeriph_GPIOE | RCC_AHBPeriph_GPIOH, ENABLE);

  /* Configure all GPIO port pins in Analog Input mode (floating input trigger OFF) */
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_All;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_40MHz;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AIN;
  
  GPIO_Init(GPIOB, &GPIO_InitStructure);
  GPIO_Init(GPIOC, &GPIO_InitStructure);
  GPIO_Init(GPIOD, &GPIO_InitStructure);
  GPIO_Init(GPIOE, &GPIO_InitStructure);
  GPIO_Init(GPIOH, &GPIO_InitStructure);

#if  DEBUG_SWD_PIN == 1
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_All & (~GPIO_Pin_13) & (~GPIO_Pin_14);
#endif
  
  //PA11 PA12 is USB pin
  GPIO_Init(GPIOA, &GPIO_InitStructure);
  
  /* Disable GPIOs clock */ 	
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA | RCC_AHBPeriph_GPIOB | 
                        RCC_AHBPeriph_GPIOC | RCC_AHBPeriph_GPIOD | 
                        RCC_AHBPeriph_GPIOE | RCC_AHBPeriph_GPIOH, DISABLE);
}

void  InitUsbPin (void)
{
  GPIO_InitTypeDef GPIO_InitStructure;

  //PC10 - is USB_ON
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);

  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_10;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_DOWN;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_Init(GPIOC, &GPIO_InitStructure);

  //PA11 PA12 is USB pin
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_11 | GPIO_Pin_12;
  GPIO_Init(GPIOA, &GPIO_InitStructure);
}  

void  DisableUsbPin (void)
{
  GPIO_InitTypeDef GPIO_InitStructure;
  //PC10 - is USB_ON
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_10;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_40MHz;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AIN;
  GPIO_Init(GPIOC, &GPIO_InitStructure);

  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_11 | GPIO_Pin_12;
  GPIO_Init(GPIOC, &GPIO_InitStructure);
}  

uint8_t IsUsbPower(void)
{
    return GPIO_ReadInputDataBit(GPIOC, GPIO_Pin_10);
}

void RtcInit(void)
{
  RTC_InitTypeDef RTC_InitStruct;
  //RTC_TimeTypeDef RTC_TimeStruct;

  RTC_StructInit(&RTC_InitStruct);
  RTC_Init(&RTC_InitStruct);

  //RTC_TimeStructInit(&RTC_TimeStruct);
  //RTC_SetTime(RTC_Format_BIN, &RTC_TimeStruct);
}

int16_t indexCurSample(RTC_TimeTypeDef* time)
{
  int16_t idx = ((int16_t)time->RTC_Hours%SAVE_INTERVAL_HOURS)*60+time->RTC_Minutes;

  idx /= SAVE_INTERVAL_HOURS;

  return idx;
}

uint8_t roundIntervalHours(uint8_t h)
{
  return (h/SAVE_INTERVAL_HOURS)*SAVE_INTERVAL_HOURS;
}

//first 4 byte of eeprom is flash offset
uint32_t getFlashPageOffset(void)
{
  return *(uint16_t*)EEPROM_START_ADDRESS;
}

void setFlashPageOffset(uint32_t offset)
{
  DATA_EEPROM_Unlock();
  DATA_EEPROM_FastProgramWord(EEPROM_START_ADDRESS, offset);
  DATA_EEPROM_Lock();
}

//to test save interval = 1 min
//normal save interval = 6 min
void saveData(RTC_TimeTypeDef* time, RTC_DateTypeDef* date)
{
  uint32_t offset;
  struct BigIntervalDataHeader* data;
  struct BigIntervalDataHeader header;
  struct DhtSample curSample;
  if(gErrorFlash)
    return;
  offset = getFlashPageOffset();
  data = (struct BigIntervalDataHeader*)(offset+FLASH_START_ARRAY);
  if(data->saveIntervalHours==0 
    || roundIntervalHours(data->hour)!=roundIntervalHours(time->RTC_Hours) 
    || data->mday!=date->RTC_Date || data->mon!=date->RTC_Month || data->year!=date->RTC_Year)
  {
    if(data->saveIntervalHours!=0)
      offset+=256;
    if(offset>=FLASH_STORAGE_SIZE)
      offset=0;
    setFlashPageOffset(offset);
    data = (struct BigIntervalDataHeader*)(offset+FLASH_START_ARRAY);

    FLASH_Unlock();
    if(FLASH_ErasePage(offset+FLASH_START_ARRAY)!=FLASH_COMPLETE)
    {
      gErrorFlash = true;
      FLASH_Lock();
      return;
    }
   
    header.saveIntervalHours = SAVE_INTERVAL_HOURS;
    header.min = 0;
    header.hour = roundIntervalHours(time->RTC_Hours);
    header.mday = date->RTC_Date;
    header.mon = date->RTC_Month;
    header.year = date->RTC_Year;
    header.padder0 = 0;
    header.padder1 = 0;

    FLASH_FastProgramWord(offset+FLASH_START_ARRAY, *(uint32_t*)&header);
    FLASH_FastProgramWord(offset+FLASH_START_ARRAY+4, *(1+(uint32_t*)&header));

    FLASH_Lock();
  }

  DhtMeasureAndDisplay();

  FLASH_Unlock();
  
  curSample.temperature = dht_temperature;
  curSample.humidity = dht_humidity;
  FLASH_FastProgramWord(offset+FLASH_START_ARRAY+sizeof(struct BigIntervalDataHeader)+
    sizeof(struct DhtSample)*indexCurSample(time), *(uint32_t*)&curSample);
  /*
  tm = time->RTC_Seconds + 0x100*(uint32_t)time->RTC_Minutes+0x10000*(uint32_t)time->RTC_Hours+0x1000000*(uint32_t)date->RTC_Date;
  FLASH_FastProgramWord(offset+FLASH_START_ARRAY+sizeof(struct BigIntervalDataHeader)+
    sizeof(struct DhtSample)*indexCurSample(time), tm);
  */
  FLASH_Lock();
}

void ClockQuant(void)
{
  int16_t curSampleIndex;
  RTC_TimeTypeDef time;
  RTC_DateTypeDef date;
  RTC_GetTime(RTC_Format_BIN, &time);
  RTC_GetDate(RTC_Format_BIN, &date);

  //time not set
  if(date.RTC_Year==0)
    return;

  //Bugfix: Not write in midnight
  if(time.RTC_Hours==23 && time.RTC_Minutes==59 && time.RTC_Seconds>55)
    return;
  if(time.RTC_Hours==0 && time.RTC_Minutes==0 && time.RTC_Seconds<5)
    return;

  curSampleIndex = indexCurSample(&time);

  if(curSampleIndex==gLastSampleIndex)
    return;

  //Bugfix: Read date twice to avoid midnight bug
  RTC_GetDate(RTC_Format_BIN, &date);
  saveData(&time, &date);

  gLastSampleIndex = curSampleIndex;
}

#ifdef  USE_FULL_ASSERT

/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t* file, uint32_t line)
{ 
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */

  /* Infinite loop */
  while (1)
  {
  }
}

#endif
